import pathlib
from ._upath import Upath
from ._local import LocalUpath

from overrides import overrides, EnforceOverrides


def _resolve_local_path(p):
    if isinstance(p, str):
        p = pathlib.Path(p)
    if isinstance(p, pathlib.Path):
        p = LocalUpath(str(p.absolute()))
    else:
        assert isinstance(p, LocalUpath)
    return p


class BlobUpath(Upath, EnforceOverrides):
    @property
    def blob_name(self) -> str:
        return self._path.lstrip('/')

    def download_dir(self, target, *, overwrite=False) -> int:
        target_ = _resolve_local_path(target)
        return self.export_dir(target_, overwrite=overwrite)

    def download_file(self, target, *, overwrite=False) -> None:
        target_ = _resolve_local_path(target)
        return self.export_file(target_, overwrite=overwrite)

    @overrides
    def is_dir(self) -> bool:
        '''In a typical blob store, there is no such concept as a
        "directory". Here we emulate the concept in a local file
        system. If there is a blob named like

            /ab/cd/ef/g.txt

        we say there exists directory "/ab/cd/ef".
        We should never have a trailing `/` in a blob's name, like

            /ab/cd/ef/

        (I don't know whether the blob stores allow
        such blob names.)

        Consequently, `is_dir` is equivalent
        to "having stuff in the dir". There is no such thing as
        an "empty directory" in blob stores.
        '''
        try:
            next(self.iterdir())
            return True
        except StopIteration:
            return False

    @overrides
    def iterdir(self):
        # A naive, inefficient implementation.
        p0 = self._path  # this could be '/'.
        if not p0.endswith('/'):
            p0 += '/'
        np0 = len(p0)
        subdirs = set()
        for p in self.riterdir():
            tail = p._path[np0:]
            if tail.startswith('/'):
                raise Exception(f"malformed blob name '{p._path}'")
            if '/' in tail:
                tail = tail[: tail.find('/')]
            if tail not in subdirs:
                yield self / tail
                subdirs.add(tail)

    def upload_dir(self, source, *, overwrite=False) -> int:
        s = _resolve_local_path(source)
        return self.import_dir(s, overwrite=overwrite)

    def upload_file(self, source, *, overwrite=False) -> None:
        s = _resolve_local_path(source)
        return self.import_file(s, overwrite=overwrite)



from __future__ import annotations
# Enable using `Upath` in type annotations in the code
# that defines this class.
# https://stackoverflow.com/a/49872353
# Will no longer be needed in Python 3.10.

import contextlib
import logging
import os
# import socket
# import urllib3
from io import BufferedReader, UnsupportedOperation, BytesIO
from typing import Union

# import requests
from google import resumable_media
from google.oauth2 import service_account
from google.cloud import storage
from google.api_core.exceptions import (
    NotFound, PreconditionFailed, TooManyRequests)
from overrides import overrides

from ._upath import FileInfo, Upath, LockAcquisitionTimeoutError
from ._blob import BlobUpath
from ._local import LocalUpath
from ._util import Backoff


logger = logging.getLogger(__name__)


# 67108864 = 256 * 1024 * 256 = 64 MB
MEGABYTES32 = 33554432
MEGABYTES64 = 67108864


class GcpBlobUpath(BlobUpath):
    def __init__(self, *paths: str, bucket_name: str, account_info: dict):
        '''
        `account_info`: a dict with these elements:
            'type': 'service_account',
            'project_id':
            'private_key_id':
            'private_key':
                '-----BEGIN PRIVATE KEY-----\n'
                + private_key.encode('latin1').decode('unicode_escape')
                + '\n-----END PRIVATE KEY-----\n',
            'client_email':
            'client_id':
            'auth_uri': 'https://accounts.google.com/o/oauth2/auth',
            'token_uri': 'https://oauth2.googleapis.com/token',
            'auth_provider_x509_cert_url': 'https://www.googleapis.com/oauth2/v1/certs',
            'client_x509_cert_url': f"https://www.googleapis.com/robot/v1/metadata/x509/{client_email.replace('@', '%40')}"
        '''
        super().__init__(*paths)
        self.bucket_name = bucket_name
        self._account_info = account_info
        self._client = None
        self._bucket = None
        self._blob = None
        self._lock_count: int = 0
        self._generation = -1

    def __repr__(self) -> str:
        return "{}('{}', bucket_name='{}')".format(
            self.__class__.__name__, self._path, self.bucket_name
        )

    def __str__(self) -> str:
        return f"{self.bucket_name}://{self._path}"

    def __eq__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other.bucket_name != self.bucket_name:
            return NotImplemented
        return self._path == other._path

    def __lt__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other.bucket_name != self.bucket_name:
            return NotImplemented
        return self._path < other._path

    def __le__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other.bucket_name != self.bucket_name:
            return NotImplemented
        return self._path <= other._path

    def __gt__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other.bucket_name != self.bucket_name:
            return NotImplemented
        return self._path > other._path

    def __ge__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other.bucket_name != self.bucket_name:
            return NotImplemented
        return self._path >= other._path

    def __getstate__(self):
        # Customize pickle because `self._client` and `self._bucket`
        # (when not None) can't be pickled.
        return {
            '_path': self._path,
            'bucket_name': self.bucket_name,
            '_account_info': self._account_info,
        }

    def __setstate__(self, data):
        self._path = data['_path']
        self.bucket_name = data['bucket_name']
        self._account_info = data['_account_info']
        self._client = None
        self._bucket = None
        self._blob = None
        self._lock_count = 0

    @property
    def client(self):
        if self._client is None:
            gcp_cred = service_account.Credentials.from_service_account_info(
                self._account_info)
            self._client = storage.Client(
                project=self._account_info['project_id'],
                credentials=gcp_cred,
            )
        return self._client

    @property
    def bucket(self):
        if self._bucket is None:
            self._bucket = self.client.bucket(self.bucket_name)
            # self._bucket = self.client.get_bucket(self.bucket_name)
        return self._bucket

    def blob(self):
        '''
        This constructs a Blob object irrespecitive of whether the blob
        exists in cloud storage.
        '''
        if self._blob is None:
            self._blob = self.bucket.blob(self.blob_name)
        return self._blob

    def get_blob(self):
        '''
        While `.blob` simply constructs a Blob object,
        `.get_blob` makes network calls to refresh properties
        of the object in cloud storage. If the blob does not exist,
        return `None`.
        '''
        b = self.blob()
        try:
            b.reload(client=self.client)
            return b
        except NotFound:
            return None

    # def _blob_retry(self, func_name, *args, max_tries=3, **kwargs):
    #     # `func_name` is the name of a blob method.
    #     sleeper = Backoff(0.1)
    #     while True:
    #         try:
    #             return getattr(self.blob(), func_name)(*args, **kwargs)
    #         except (requests.exceptions.ReadTimeout,
    #                 urllib3.exceptions.ReadTimeoutError,
    #                 socket.timeout,
    #                 urllib3.exceptions.SSLError,
    #                 requests.exceptions.SSLError,
    #                 requests.exceptions.ConnectionError,
    #                 ) as e:
    #             if sleeper.retries >= max_tries:
    #                 raise
    #             logger.info("retrying %r on error: %r", func_name, e)
    #             sleeper.sleep()
    #             # My hypothesis is that sometimes timeout happens
    #             # because the cached `client`, `bucket`, `blob`
    #             # have become stale, hence should be recreated.
    #             self._client = None
    #             self._bucket = None
    #             self._blob = None

    def _blob_rate_limit(self, func, *args, max_retries=5, **kwargs):
        # `func_name` is the name of a create/update/delete function.
        sleeper = Backoff(0.2)
        while True:
            try:
                return func(*args, **kwargs)
            except TooManyRequests as e:
                if sleeper.retries >= max_retries:
                    raise
                logger.info("retrying %r on error: %r", func, e)
                sleeper.sleep()

    @overrides
    def _copy_file(self, target: GcpBlobUpath, *, overwrite=False) -> None:
        # https://cloud.google.com/storage/docs/copying-renaming-moving-objects
        try:
            self.bucket.copy_blob(
                self.blob(), target.bucket, target.blob_name, client=self.client,
                if_generation_match=None if overwrite else 0,
            )
        except NotFound:
            raise FileNotFoundError(self)
        except PreconditionFailed:
            raise FileExistsError(target)

    @overrides
    def _export_file(self, target: Upath, *, overwrite=False) -> None:
        if not isinstance(target, LocalUpath):
            return super()._export_file(target, overwrite=overwrite)
        if not overwrite and target.is_file():
            raise FileExistsError(target)
        os.makedirs(str(target.parent), exist_ok=True)
        try:
            with open(target.localpath, 'wb') as file_obj:
                self._read_into_buffer(file_obj)
            updated = self.blob().updated
            if updated is not None:
                mtime = updated.timestamp()
                os.utime(target.localpath, (mtime, mtime))
        except resumable_media.DataCorruption:
            target.remove_file()
            raise

    @overrides
    def file_info(self):
        b = self.get_blob()
        if not b:
            return None
        return FileInfo(
            ctime=b.time_created.timestamp(),
            mtime=b.updated.timestamp(),
            time_created=b.time_created,
            time_modified=b.updated,
            size=b.size,  # bytes
            details=b._properties,
        )
        # If an existing file is written to again using `write_...`,
        # then its `ctime` and `mtime` are both updated.
        # My experiments showed that `ctime` and `mtime` are equal.

    @overrides
    def _import_file(self, source: Upath, *, overwrite=False) -> None:
        if not isinstance(source, LocalUpath):
            return super()._import_file(source, overwrite=overwrite)

        filename = str(source)
        content_type = self.blob()._get_content_type(None, filename=filename)

        def _upload():
            with open(filename, 'rb') as file_obj:
                total_bytes = os.fstat(file_obj.fileno()).st_size
                self._write_from_buffer(
                    file_obj,
                    size=total_bytes,
                    content_type=content_type,
                    overwrite=overwrite,
                )

        self._blob_rate_limit(_upload)

    @overrides
    def is_file(self) -> bool:
        # This is not cached, in case the object is modified anytime
        # by other clients.
        return self.blob().exists(self.client)

    @overrides
    def iterdir(self):
        # From Google doc:
        #
        # Lists all the blobs in the bucket that begin with the prefix.
        #
        # This can be used to list all blobs in a "folder", e.g. "public/".
        #
        # The delimiter argument can be used to restrict the results to only the
        # "files" in the given "folder". Without the delimiter, the entire tree under
        # the prefix is returned. For example, given these blobs:
        #
        #     a/1.txt
        #     a/b/2.txt
        #
        # If you specify prefix ='a/', without a delimiter, you'll get back:
        #
        #     a/1.txt
        #     a/b/2.txt
        #
        # However, if you specify prefix='a/' and delimiter='/', you'll get back
        # only the file directly under 'a/':
        #
        #     a/1.txt
        #
        # As part of the response, you'll also get back a blobs.prefixes entity
        # that lists the "subfolders" under `a/`:
        #
        #     a/b/
        #
        # Search "List the objects in a bucket using a prefix filter | Cloud Storage"

        prefix = self.blob_name + '/'
        k = len(prefix)
        blobs = self.client.list_blobs(self.bucket, prefix=prefix, delimiter='/')
        for p in blobs:
            obj = self / p.name[k:]  # files
            obj._blob = p
            yield obj
        for p in blobs.prefixes:
            yield self / p[k:].rstrip('/')  # "subdirectories"

    def _acquire_lease(self, *, timeout: int = None):
        if self._path == '/':
            raise UnsupportedOperation("can not write to root as a blob", self)
        if timeout is None:
            timeout = 600  # seconds
        sleeper = Backoff(0.1)
        while True:
            try:
                # b.upload_from_string(b'0', if_generation_match=0)
                # b.cache_control = 'no-store'
                # b.patch()
                self._blob_rate_limit(self._write_bytes, b'0')
                self._generation = self.blob().generation
                return
            except (PreconditionFailed, FileExistsError):
                if (t := sleeper.time_elapsed) >= timeout:
                    raise LockAcquisitionTimeoutError(self, t)
                sleeper.sleep()

    @contextlib.contextmanager
    @overrides
    def lock(self, *, timeout=None):
        # References:
        # https://www.joyfulbikeshedding.com/blog/2021-05-19-robust-distributed-locking-algorithm-based-on-google-cloud-storage.html
        # https://cloud.google.com/storage/docs/generations-preconditions
        # https://cloud.google.com/storage/docs/gsutil/addlhelp/ObjectVersioningandConcurrencyControl

        # This implementation does not prevent the file from being deleted
        # by other workers that does not use the 'if-generation-match' condition.
        # It relies on the assumption that this blob
        # is used solely in this locking logic.

        if self._lock_count == 0:
            self._acquire_lease(timeout=timeout)
        self._lock_count += 1
        try:
            yield
        finally:
            self._lock_count -= 1
            if self._lock_count == 0:
                try:
                    self._blob_rate_limit(
                        self.blob().delete,
                        client=self.client,
                        if_generation_match=self._generation,
                        max_retries=10,
                    )
                except Exception as e:
                    logger.error(e)

    def open(self, mode='r', **kwargs):
        '''
        Use this on a blob (not a "directory") as a context manager.
        See Google documentation.
        '''
        return self.blob().open(mode, **kwargs)

    def _read_into_buffer(self, file_obj):
        file_info = self.file_info()
        if not file_info:
            raise FileNotFoundError(self)
        file_size = file_info.size  # bytes
        if file_size <= MEGABYTES32:
            try:
                self.blob().download_to_file(file_obj, client=self.client)
                return
            except NotFound:
                raise FileNotFoundError(self)

        def _download(client, blob, start, end):
            buffer = BytesIO()
            try:
                blob.download_to_file(buffer, client=client, start=start, end=end)
            except NotFound:
                raise FileNotFoundError(blob.name)
            # Both `start` and `end` are inclusive.
            # The very first `start` should be 0.
            buffer.seek(0)
            return buffer, end - start + 1

        def _do_download():
            client = self.client
            blob = self.blob()
            name = self.name
            k = 0
            p = 0
            while True:
                kk = min(k + MEGABYTES32, file_size)
                p += 1
                yield (_download, (client, blob, k, kk - 1), {}, f"downloading {name} part {p}")
                k = kk
                if k >= file_size:
                    break

        for buf, k in self._run_in_executor(_do_download()):
            n = file_obj.write(buf.getbuffer())
            if n != k:
                raise BufferError(f"expecting to read {k} bytes; actually read {n} bytes")
            buf.close()

    @overrides
    def read_bytes(self) -> bytes:
        buffer = BytesIO()
        self._read_into_buffer(buffer)
        return buffer.getvalue()

    @overrides
    def remove_file(self):
        try:
            self._blob_rate_limit(self.blob().delete, client=self.client)
            self._blob = None
        except NotFound:
            self._blob = None
            raise FileNotFoundError(self)

    @overrides
    def riterdir(self):
        prefix = self.blob_name + '/'
        k = len(prefix)
        for p in self.client.list_blobs(self.bucket, prefix=prefix):
            obj = self / p.name[k:]
            obj._blob = p
            yield obj

    @overrides
    def with_path(self, *paths: str):
        obj = self.__class__(*paths, bucket_name=self.bucket_name,
                             account_info=self._account_info)
        obj._client = self._client
        obj._bucket = self._bucket
        return obj

    def _write_from_buffer(self, file_obj, *, overwrite=False, content_type=None, size=None):
        if self._path == '/':
            raise UnsupportedOperation("can not write to root as a blob", self)

        try:
            self.blob().upload_from_file(
                file_obj,
                content_type=content_type,
                size=size,
                client=self.client,
                if_generation_match=None if overwrite else 0,
            )
            # TODO: set "create_time", 'update_time" to be the same
            # as the source local file?
            # Blob objects has methods `_set_properties`, `_patch_property`,
            # `patch`.
        except PreconditionFailed:
            raise FileExistsError(self)

    def _write_bytes(self, data, **kwargs):
        b = BytesIO(data)
        b.seek(0)
        self._write_from_buffer(b, content_type='text/plain', size=len(data), **kwargs)

    @overrides
    def write_bytes(self, data: Union[bytes, BufferedReader], *, overwrite=False):
        if isinstance(data, bytes):
            self._blob_rate_limit(self._write_bytes, data, overwrite=overwrite)
            return
        self._write_from_buffer(data, content_type='text/plain', overwrite=overwrite)



from __future__ import annotations
# Enable using `Upath` in type annotations in the code
# that defines this class.
# https://stackoverflow.com/a/49872353
# Will no longer be needed in Python 3.10.

import abc
import concurrent.futures
import contextlib
import datetime
import logging
import os
import os.path
import pathlib
import queue
import threading
from dataclasses import dataclass
from io import UnsupportedOperation
from typing import List, Iterable, Iterator, Type, TypeVar, Any, Optional, Tuple, Callable

from overrides import EnforceOverrides
from tqdm import tqdm
from .serializer import (
    ByteSerializer, TextSerializer,
    JsonSerializer, ZJsonSerializer, ZstdJsonSerializer,
    PickleSerializer, ZPickleSerializer, ZstdPickleSerializer,
    OrjsonSerializer, ZOrjsonSerializer, ZstdOrjsonSerializer,
)


logger = logging.getLogger(__name__)
T = TypeVar('T', bound='Upath')


class LockAcquisitionTimeoutError(TimeoutError):
    pass


@dataclass
class FileInfo:
    ctime: float   # creation POSIX timetamp
    mtime: float   # last modification POSIX timestamp
    time_created: datetime.datetime
    time_modified: datetime.datetime
    size: int      # in bytes
    details: Any   # platform-dependent


class Upath(abc.ABC, EnforceOverrides):  # pylint: disable=too-many-public-methods
    NUM_EXECUTORS = 16  # TODO: what is a good value in general?
    _thread_executor_ = None

    @classmethod
    def _run_in_executor(cls, tasks: Iterable[Tuple[Callable, tuple, dict, str]]):
        '''
        This method is used to run multiple I/O jobs concurrently, e.g.
        uploading/downloading all files in a folder recursively.
        Where supported, this may also be used to download a large blob
        by splitting the work into multiple segments.

        `tasks`: each element is a tuple of (func, args, kwargs, description).
        '''
        if not isinstance(tasks, list):
            tasks = list(tasks)
        n_tasks = len(tasks)
        if not n_tasks:
            return

        if cls._thread_executor_ is None:
            cls._thread_executor_ = concurrent.futures.ThreadPoolExecutor(cls.NUM_EXECUTORS)

        def enqueue(tasks, executor, q):
            for func, args, kwargs, desc in tasks:
                t = executor.submit(func, *args, **kwargs)
                q.put((t, desc))
            q.put(None)

        q = queue.Queue(cls.NUM_EXECUTORS * 2)
        task = threading.Thread(target=enqueue, args=(tasks, cls._thread_executor_, q))
        task.start()
        with tqdm(total=n_tasks) as pbar:
            while True:
                z = q.get()
                if z is None:
                    break
                t, desc = z
                pbar.set_description(desc + '... ...')
                pbar.update(0.5)
                try:
                    yield t.result()
                except Exception:
                    while True:
                        z = q.get()
                        if z is None:
                            break
                        t, desc = z
                        t.cancel()
                        # This may not succeed, but there isn't a good way to
                        # guarantee cancellation here.
                    raise
                pbar.set_description(desc + '... ... done')
                pbar.update(0.5)
            task.join()

    @classmethod
    def register_read_write_byte_format(cls, serde: Type[ByteSerializer], name: str):
        '''
        For example, if `serde` is a ByteSerializer subclass and `name` is 'myway',
        then this method adds isinstance methods `write_myway` and `read_myway`.

        `name`: needs to be valid method name, e.g. can't contain space or dash.
        '''
        def _write(self, data, *, overwrite=False, **kwargs):
            return self.write_bytes(
                serde.serialize(data, **kwargs),
                overwrite=overwrite)

        def _read(self, **kwargs):
            z = self.read_bytes()
            return serde.deserialize(z, **kwargs)

        setattr(_write, '__name__', f'write_{name}')
        setattr(_read, '__name__', f'read_{name}')
        setattr(cls, f'write_{name}', _write)
        setattr(cls, f'read_{name}', _read)

    @classmethod
    def register_read_write_text_format(cls, serde: Type[TextSerializer], name: str):
        '''
        Anologous to `register_read_write_byte_format`.
        '''
        def _write(self, data, *, overwrite=False, **kwargs):
            return self.write_text(
                serde.serialize(data, **kwargs),
                overwrite=overwrite)

        def _read(self, **kwargs):
            z = self.read_text()
            return serde.deserialize(z, **kwargs)

        setattr(_write, '__name__', f'write_{name}')
        setattr(_read, '__name__', f'read_{name}')
        setattr(cls, f'write_{name}', _write)
        setattr(cls, f'read_{name}', _read)

    def __init__(self, *pathsegments: str):
        '''`Upath` is the base class for a client to a blob store,
        including local file system as a special case.

        `*pathsegments`: analogous to the input to `pathlib.Path`.
        The first segment may or may not start with `/`; it makes
        no difference. The path constructed with `*pathsegments`
        is always "absolute" under a known "root".

            Note that if one segment starts with '/', it will reset to the "root"
            and discard all the segments that have come before it.

            If missing, the path constructed is the "root".

        For a local POSIX file system, the root is the regular `/`.
        For Azure blob store, the root is that of a "container".
        For AWS and GCP blob stores, the root is that of a "bucket".
        '''

        self._path = os.path.normpath(os.path.join(
            '/', *pathsegments))  # pylint: disable=no-value-for-parameter
        # The path is always "absolute" starting with '/'.
        # Unless the path is just `/` itself, it does not have a trailing `/`.

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self._path}')"
        # Subclass may want to customize this method to add more info,
        # e.g. "bucket" name.

    def __str__(self) -> str:
        return self._path

    def __eq__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self._path == other._path
        # Subclass may want to customize this method to check more things,
        # e.g. whether `self` and `other` are in the same "bucket".

    def __lt__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self._path < other._path

    def __le__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self._path <= other._path

    def __gt__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self._path > other._path

    def __ge__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self._path >= other._path

    def __hash__(self) -> int:
        return hash(repr(self))

    def __truediv__(self: T, key: str) -> T:
        '''
        This is called by `self / key`.
        '''
        return self.joinpath(key)

    def copy_dir(self, target: str, *, overwrite: bool = False) -> int:
        '''Analogous to `copy_file`.

        Return the number of files copied.
        '''
        target_ = self.parent / target
        if target_ == self:
            return 0

        def foo():
            for p in self.riterdir():
                extra = str(p.path.relative_to(self.path))
                yield (
                    p.copy_file,
                    ((target_ / extra)._path, ),
                    {'overwrite': overwrite},
                    f'copying {p.name}',
                )

        n = 0
        for _ in self._run_in_executor(foo()):
            n += 1
        return n

    def _copy_file(self: T, target: T, *, overwrite: bool = False) -> None:
        # `target` is a path in the same store.
        # Reference implementation.
        # Subclass may customize this to perform file operations.
        target.write_bytes(self.read_bytes(), overwrite=overwrite)

    def copy_file(self, target: str, *, overwrite: bool = False) -> None:
        '''Copy file to `target` in the same store.

        `target` is either absolute, or relative to `self.parent`.
        For example, if `self` is '/a/b/c/d.txt', then
        `target='e.txt'` means '/a/b/c/e.txt'.

        If `self` is not an existing file, raise `FileNotFoundError`.

        If `target` is an existing file and `overwrite` is `False`,
        raise `FileExistsError`.

        If `target` is an existing directory, raise `IsADirectoryError`.
        Note: this behavior is different from the Unix command `cp`
        in this situation---it does not *copy into* the target directory.
        '''
        target_ = self.parent / target
        if target_ == self:
            return

        if target_.is_dir():
            raise IsADirectoryError(target_)

        self._copy_file(target_, overwrite=overwrite)

    def exists(self) -> bool:
        '''Return `True` if the path is an existing file or dir,
        `False` otherwise.

        In a blobstore with blobs

            /a/b/cd
            /a/b/cd/e.txt

        '/a/b/cd' exists, and is both a file and a dir;
        '/a/b/cd/e.txt' exists, and is a file;
        '/a/b' exists, and is a dir;
        '/a/b/c' does not exist.
        '''
        return self.is_file() or self.is_dir()

    def export_dir(self, target: Upath, *, overwrite: bool = False) -> int:
        '''Copy the content of the current directory recursively
        to the specified `target`, which is typically in another store.

        `target` corresponds to `self`, that is, direct children of `self`
        are copied directly into `target`.

        Compare with `copy_dir`, which make copies within the same store.

        Overwriting happens file-wise. For example, if the target directory
        contains files that do not exist in the source directory, they
        are left untouched.

        If `target` is a `LocalUpath` object, then a subclass may implement more
        efficient ways to "download", and also renaming this method to "download_dir".

        Return the number of files copied.
        '''
        def foo():
            self_path = self.path
            for p in self.riterdir():
                extra = str(p.path.relative_to(self_path))
                yield (
                    p.export_file,
                    (target / extra, ),
                    {'overwrite': overwrite},
                    f'exporting {p.name}',
                )

        n = 0
        for _ in self._run_in_executor(foo()):
            n += 1
        return n

    def _export_file(self, target: Upath, *, overwrite: bool = False) -> None:
        # Reference implementation.
        # Subclass may customize this to perform file download
        # when `target` is a `LocalUpath`.
        target.write_bytes(self.read_bytes(), overwrite=overwrite)

    def export_file(self, target: Upath, *, overwrite: bool = False) -> None:
        '''Copy the file to the specified `target`, which is typically
        in another store.

        The `target` specifies a blob corresponding to `self`.
        If `target` is an existing directory, a `IsADirectoryError` is raised.
        A copy is not placed *into* the target directory. This behavior
        differs from the Linux command `cp`.

        If `target` is a `LocalUpath` object, then a subclass may implement more
        efficient ways to "download", and also renaming this method to "download_file".

        Compare with `copy_file`, which makes copies within the same store.
        '''
        if target.is_dir():
            # Do not delete.
            raise IsADirectoryError(target)

        self._export_file(target, overwrite=overwrite)

    @ abc.abstractmethod
    def file_info(self) -> Optional[FileInfo]:
        '''
        If `self.is_file()` is `False`, return `None`.
        '''
        raise NotImplementedError

    def import_dir(self, source: Upath, *, overwrite: bool = False) -> int:
        '''Analogous to `export_dir`.
        '''
        def foo():
            source_path = source.path
            for p in source.riterdir():
                extra = str(p.path.relative_to(source_path))
                yield (
                    (self / extra).import_file,
                    (p, ),
                    {'overwrite': overwrite},
                    f'importing {p.name}',
                )

        n = 0
        for _ in self._run_in_executor(foo()):
            n += 1
        return n

    def _import_file(self, source: Upath, *, overwrite: bool = False) -> None:
        self.write_bytes(source.read_bytes(), overwrite=overwrite)

    def import_file(self, source: Upath, *, overwrite: bool = False) -> None:
        # When `target` is a `LocalUpath`, subclass may implement this
        # in more efficient ways for uploading, and rename it to `upload_file`.
        if self.is_dir():
            # Do not delete.
            raise IsADirectoryError(self)
        self._import_file(source, overwrite=overwrite)

    @ abc.abstractmethod
    def is_dir(self) -> bool:
        '''Return `True` if the path is an existing directory, `False` otherwise.

        If there exists a file named like

            /a/b/c/d.txt

        we say `/a`, `/a/b`, `/a/b/c` are existing directories.

        In a cloud blob store, there's no such thing as an
        "empty directory", because there is no concept of "directory".
        A blob store just consists of files (aka blobs) with names,
        which could contain the letter '/', with no special meaning
        attached to it.
        We interpret the name `/a/b` as a directory
        to emulate the familiar concept in a local file system because
        there exist files named `/a/b/...`.

        In a local file system, there can be empty directories.
        However, it is recommended to not have empty directories.

        There is no method for "creating a dir" (like `mkdir`).
        Simply create a file under the dir, and the dir will come into being.
        This is analogous to our treatment to files---we don't "create" a file
        in advance; we simply write to a path, intending it to be a file.
        '''
        raise NotImplementedError

    @ abc.abstractmethod
    def is_file(self) -> bool:
        '''Return `True` if the path is an existing file, `False` otherwise.

        In a cloud blob store, a path can be both a file and a dir. For
        example, if these blobs exist:

            /a/b/c/d.txt
            /a/b/c

        we say `/a/b/c` is a "file", and also a "dir".
        User is recommended to avoid such namings.

        This situation does not happen in a local file system.
        '''
        raise NotImplementedError

    @ abc.abstractmethod
    def iterdir(self: T) -> Iterator[T]:
        '''Yield the first-level (i.e. non-recursive) children
        of the current dir.

        Each yielded element is either a file or a dir.

        If `self` is not a dir (e.g. maybe it's a file),
        or does not exist at all, yield nothing (resulting in an
        empty iterable), but do not raise exception.

        There is no guarantee on the order of the returned elements.
        '''
        raise NotImplementedError

    def joinpath(self: T, *other: str) -> T:
        '''Join this path with more segments, return the new path object.'''
        return self.with_path(self._path, *other)

    @ contextlib.contextmanager
    @ abc.abstractmethod
    def lock(self, *, timeout: int = None):
        '''Lock the file pointed to, in order to have exclusive access.

        `timeout`: if the lock can't be acquired within *timeout* seconds,
        raise `LockAcquisitionTimeoutError`. If `None`, wait for ever until
        a lock is acquired. Once a lease is acquired,
        it will not expire until this contexmanager exits.
        In other words, this is timeout for the "wait", not for the
        lease itself. Actual waiting time may be slightly longer.

        This is a "mandatory lock", as opposed to an "advisory lock".
        However, this API does not specify that the locked file
        can be accessed for its content or used in any particular way.
        The intended use case is for this lock to be used
        for implementing a (cooperative) "code lock".

        The `yield` statement is not required to yield anything,
        that is, it may be simply

            yield

        rather than, say,

            yield self

        One way to achive cooperative locking on a file via this mandatory
        lock is like this:

            f = Upath('abc.txt')
            with f.with_suffix('.txt.lock').lock():
                ...
                # now write `f` with exclusive access,
                # because any other (cooperative) code block
                # will not be able to get hold of `abc.txt.lock`
                # in order to write `f` in its context-managed block.
                # It's up to the program design whether this lock
                # covers reading as well.

        Some storage engines may not provide the capability to implement
        this lock.
        '''
        raise NotImplementedError

    def ls(self: T) -> List[T]:
        return sorted(self.iterdir())

    @ property
    def name(self) -> str:
        '''Return the segment after the last `/`.

        If `self.path` is '/', then `self.path.name` is ''.
        '''
        return self.path.name

    @ property
    def parent(self: T) -> T:
        return self.with_path(str(self.path.parent))

    @ property
    def path(self) -> pathlib.PurePosixPath:
        return pathlib.PurePosixPath(self._path)

    @ abc.abstractmethod
    def read_bytes(self) -> bytes:
        '''Return the binary contents of the file.

        If `self` is not a file or is non-existent,
        raise `FileNotFoundError`.
        '''
        raise NotImplementedError

    def read_text(self):
        # Refer to https://docs.python.org/3/library/functions.html#open
        return self.read_bytes().decode(encoding='utf-8', errors='strict')

    # TODO: rename 'remove' to 'delete'?

    def remove_dir(self) -> int:
        '''Remove the directory pointed to by `self`,
        along with all its contents, recursively.

        Return the number of files removed.

        Local upath needs to customize this implementation, because
        it needs to take care of deleting "empty" subdirectories.
        '''
        def foo():
            for p in self.riterdir():
                yield p.remove_file, [], {}, f'deleting {p.name}'

        n = 0
        for _ in self._run_in_executor(foo()):
            n += 1
        return n

    @abc.abstractmethod
    def remove_file(self) -> None:
        '''Remove the file pointed to by `self`.

        If `self` is not an existing file, raise FileNotFoundError.
        If the file exists but can't be removed, usually the platform-dependent
        exception is propagated.
        '''
        raise NotImplementedError

    def rename_dir(self: T, target: str, *, overwrite: bool = False) -> T:
        '''Analogous to `rename_file`.

        `overwrite` is applied per file, which suggests that if there are
        files under `target` that do not have counterparts under `self`,
        they are left untouched.

        Local upath needs to customize this implementation, because
        it needs to take care to delete empty subdirectories under `self`.
        '''
        if not self.is_dir():
            raise FileNotFoundError(str(self))

        target_ = self.parent / target
        if target_ == self:
            return self

        def foo():
            for p in self.riterdir():
                extra = str(p.path.relative_to(self.path))
                yield (
                    p.rename_file,
                    [(target_ / extra)._path],
                    {'overwrite': overwrite},
                    f'renaming {p.name}',
                )

        for _ in self._run_in_executor(foo()):
            pass
        return target_

    def _rename_file(self: T, target: str, *, overwrite: bool = False):
        '''Rename `self` to `target`, which is a path in the same store.

        This is a reference implementation. There are likely
        more efficient ways to do this on any specific platform.
        '''
        self.copy_file(target, overwrite=overwrite)
        self.remove_file()

    def rename_file(self: T, target: str, *, overwrite: bool = False) -> T:
        '''Rename the current file to `target` in the same store.

        `target` is either absolute or relative to `self.parent`.
        For example, if `self` is '/a/b/c/d.txt', then
        `target='e.txt'` means '/a/b/c/e.txt'.

        Return an object pointing to the new path.
        '''
        target_ = self.parent / target
        if target_ == self:
            return self

        self._rename_file(target_._path, overwrite=overwrite)
        return target_

    @abc.abstractmethod
    def riterdir(self: T) -> Iterator[T]:
        '''Yield files under the current dir recursively.

        Compared to `iterdir`, this is recursive, and yields
        *files* only. Empty subdirectories will have no representation
        in the return.

        Similar to `iterdir`, if `self` is not a dir or does not exist,
        then nothing is yielded, and no exception is raised either.

        There is no guarantee on the order of the returned elements.
        '''
        raise NotImplementedError

    def rmrf(self) -> int:
        '''Analogous to `rm -rf`. Remove the file or dir pointed to
        by `self`.

        Return the number of files removed.

        For example, if these blobs are present:

            /a/b/c/d/e.txt
            /a/b/c/kk.data
            /a/b/c

        then `Upath('/a/b/c').rmrf()` would remove all of them.
        '''
        if self._path == '/':
            raise UnsupportedOperation("`rmrf` not allowed on root directory")
        try:
            self.remove_file()
        except (FileNotFoundError, IsADirectoryError):
            n = 0
        else:
            n = 1
        try:
            m = self.remove_dir()
        except FileNotFoundError:
            m = 0
        return n + m

    @ property
    def stem(self) -> str:
        return self.path.stem

    @ property
    def suffix(self) -> str:
        return self.path.suffix

    @ property
    def suffixes(self) -> List[str]:
        return self.path.suffixes

    def with_name(self: T, name: str) -> T:
        return self.with_path(str(self.path.with_name(name)))

    def with_path(self: T, *paths) -> T:
        '''
        Returns a new object of the same class at the specified path.
        The new path is unrelated to the current path; in other words,
        the new path is not "relative" to the current path.

        Meta data such as account info remains the same.

        Subclass needs to reimplement this method
        if its `__init__` expects additional args.
        '''
        return self.__class__(*paths)

    # def with_stem(self: T, stem: str) -> T:
    #     # Available in Python 3.9+.
    #     return self.with_path(str(self.path.with_stem(stem)))

    def with_suffix(self: T, suffix: str) -> T:
        '''`suffix` should include a dot, like '.txt'.
        If `suffix` is '', the effect is to remove the existing suffix.
        '''
        return self.with_path(str(self.path.with_suffix(suffix)))

    @ abc.abstractmethod
    def write_bytes(self,
                    data: bytes,
                    *,
                    overwrite: bool = False) -> None:
        '''Write bytes to file.

        Parent directories are created as needed.

        `overwrite`: overwrite existing file?
            If `False`, and file exists, raises `FileExistsError`.
        '''
        raise NotImplementedError

    def write_text(self,
                   data: str,
                   *,
                   overwrite: bool = False,
                   ) -> None:
        z = data.encode(encoding='utf-8', errors='strict')
        self.write_bytes(z, overwrite=overwrite)


# Add methods
# 'read_json', 'write_json', 'read_json_z', 'write_json_z', 'read_json_zstd', 'write_json_zstd',
# 'read_pickle', 'write_pickle', 'read_pickle_z', 'write_pickle_z', 'read_pickle_zstd', 'write_pickle_zstd',
# 'read_orjson', 'write_orjson', 'read_orjson_z', 'write_orjson_z', 'read_orjson_zstd', 'write_orjson_zstd',
#
# Applications can follow these examples to define and register their custom formats.

Upath.register_read_write_text_format(JsonSerializer, 'json')
Upath.register_read_write_byte_format(ZJsonSerializer, 'json_z')
Upath.register_read_write_byte_format(ZstdJsonSerializer, 'json_zstd')
Upath.register_read_write_byte_format(PickleSerializer, 'pickle')
Upath.register_read_write_byte_format(ZPickleSerializer, 'pickle_z')
Upath.register_read_write_byte_format(ZstdPickleSerializer, 'pickle_zstd')
Upath.register_read_write_byte_format(OrjsonSerializer, 'orjson')
Upath.register_read_write_byte_format(ZOrjsonSerializer, 'orjson_z')
Upath.register_read_write_byte_format(ZstdOrjsonSerializer, 'orjson_zstd')



import random
import time


class Backoff:
    def __init__(self, base=1, jitter=None, multiplier=2):
        self._base = base
        self._jitter = base if jitter is None else jitter
        self._multiplier = multiplier
        self._time_started = time.perf_counter()
        self.retries = 0

    def sleep(self):
        t = self._base * self._multiplier ** self.retries + random.uniform(0, self._jitter)
        time.sleep(t)
        self.retries += 1

    @property
    def time_elapsed(self):
        return time.perf_counter() - self._time_started  # seconds



from __future__ import annotations
# Enable using `Upath` in type annotations in the code
# that defines this class.
# https://stackoverflow.com/a/49872353
# Will no longer be needed in Python 3.10.

import logging
import os
import random
import time
from contextlib import contextmanager
from datetime import datetime
from io import UnsupportedOperation

from azure.storage.blob import ContainerClient, BlobClient, BlobLeaseClient
# from azure.storage.blob.aio import (
# ContainerClient as aContainerClient,
# BlobClient as aBlobClient,
# BlobLeaseClient as aBlobLeaseClient,
# )
from azure.core.exceptions import ResourceNotFoundError, ResourceExistsError, HttpResponseError
from overrides import overrides

from ._upath import LockAcquisitionTimeoutError, FileInfo, Upath
from ._blob import BlobUpath
from ._local import LocalUpath

logging.getLogger('azure.storage').setLevel(logging.WARNING)
logging.getLogger('azure.core.pipeline.policies').setLevel(logging.WARNING)

logger = logging.getLogger(__name__)


class AzureBlobUpath(BlobUpath):
    def __init__(self,
                 *paths: str,
                 account_name: str,
                 account_key: str,
                 sas_token: str,
                 container_name: str,
                 ):
        super().__init__(*paths)

        self._account_name = account_name
        self._account_key = account_key
        self._sas_token = sas_token
        self._account_url = f"https://{account_name}.blob.core.windows.net"
        self._container_name = container_name

        self._container_client: ContainerClient = None
        self._blob_client: BlobClient = None
        # self._a_container_client: Optional[aContainerClient] = None
        # self._a_blob_client: Optional[aBlobClient] = None
        self._lease: BlobLeaseClient = None
        self._lock_count: int = 0

    def __repr__(self) -> str:
        return "{}('{}', container_name='{}')".format(
            self.__class__.__name__, self._path, self._container_name
        )

    def __str__(self) -> str:
        return f"{self._container_name}://{self._path}"

    def __eq__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other._container_name != self._container_name:
            return NotImplemented
        return self._path == other._path

    def __lt__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other._container_name != self._container_name:
            return NotImplemented
        return self._path < other._path

    def __le__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other._container_name != self._container_name:
            return NotImplemented
        return self._path <= other._path

    def __gt__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other._container_name != self._container_name:
            return NotImplemented
        return self._path > other._path

    def __ge__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return NotImplemented
        if other._container_name != self._container_name:
            return NotImplemented
        return self._path >= other._path

    @property
    def container_name(self):
        return self._container_name

    def _copy_file_from(self, source, *, overwrite=False):
        # TODO: use `overwrite`
        with self._provide_blob_client():
            with source._provide_blob_client():
                copy = self._blob_client.start_copy_from_url(
                    source._blob_client.url,
                    requires_sync=True,
                )
                assert copy['copy_status'] == 'success'

    @overrides
    def _copy_file(self, target: AzureBlobUpath, *, overwrite=False):
        target._copy_file_from(self, overwrite=overwrite)

    @overrides
    def _export_file(self, target: Upath, *, overwrite=False) -> None:
        if not isinstance(target, LocalUpath):
            return super()._export_file(target, overwrite=overwrite)
        with self._provide_blob_client():
            # TODO: check behavior of `download_blob` about
            # overwrite.
            os.makedirs(str(target.parent), exist_ok=True)
            with open(str(target), 'wb') as f:
                data = self._blob_client.download_blob()
                data.readinto(f)

    @overrides
    def file_info(self):
        try:
            with self._provide_blob_client():
                info = self._blob_client.get_blob_properties()
                return FileInfo(
                    ctime=info.creation_time.timestamp(),
                    mtime=info.last_modified.timestamp(),
                    time_created=info.creation_time,
                    time_modified=info.last_modified,
                    size=info.size,
                    details=info,
                )
                # If an existing file is written to again using
                # `write_...(..., overwrite=True)`,
                # then its `ctime` will not change; only `mtime`
                # will be updated.
        except ResourceNotFoundError:
            return None

    @overrides
    def _import_file(self, source: Upath, *, overwrite: bool = False):
        if not isinstance(source, LocalUpath):
            return super()._import_file(source, overwrite=overwrite)
        if not overwrite and self.is_file():
            # TODO: check the behavior of `upload_blob` related to
            # behavior about overwrite.
            raise FileExistsError(self)
        with self._provide_blob_client():
            with open(str(source), 'rb') as data:
                self._blob_client.upload_blob(data)

    @overrides
    def is_file(self) -> bool:
        with self._provide_blob_client():
            return self._blob_client.exists()

    @overrides
    def iterdir(self):
        with self._provide_container_client():
            prefix = self.blob_name + '/'
            k = len(prefix)
            for p in self._container_client.walk_blobs(
                    name_starts_with=prefix):
                yield self / p.name[k:]

    def _acquire_lease(self, timeout: int = None):
        if timeout is None:
            timeout = 3600
        t0 = time.perf_counter()
        while True:
            try:
                self.write_text(
                    datetime.utcnow().isoformat(), overwrite=True)
                try:
                    self._lease = self._blob_client.acquire_lease(
                        lease_duration=-1,
                        timeout=10)
                    return
                except ResourceNotFoundError:
                    continue  # go to the outer looper to write the file again
                except HttpResponseError as e:
                    if e.status_code == 409 and e.error_code == 'LeaseAlreadyPresent':  # pylint: disable=no-member
                        # Having a lease held by others. Continue to wait.
                        # This may happen when another client placed the lease
                        # on this blob right after we've created it, that is,
                        # another worker's won out in `acquire_lease`.
                        pass
                    else:
                        raise
            except HttpResponseError as e:
                if e.status_code == 412 and e.error_code == 'LeaseIdMissing':  # pylint: disable=no-member
                    # Blob exists and has a lease on it. Wait and try again.
                    pass
                else:
                    raise

            t1 = time.perf_counter()
            if t1 - t0 >= timeout:
                raise LockAcquisitionTimeoutError(self, t1 - t0)
            time.sleep(random.uniform(0.05, 1.0))

    @contextmanager
    @overrides
    def lock(self, *, timeout=None):
        '''
        References:
        https://docs.microsoft.com/en-us/azure/storage/blobs/concurrency-manage?tabs=dotnet
        '''
        with self._provide_blob_client():
            if self._lease is None:
                self._acquire_lease(timeout)
                self._lock_count = 1
            else:
                self._lock_count += 1
            try:
                yield
            finally:
                self._lock_count -= 1
                if self._lock_count <= 0:
                    self._lease.release()
                    self._lease = None
                    self._lock_count = 0

    # @asynccontextmanager
    # async def a_lock(self, *, timeout=None):
    #     loop = asyncio.get_running_loop()
    #     with self._provide_blob_client():
    #         # TODO: this context manager is sync

    #         if self._lease is None:
    #             ff = functools.partial(self._acquire_lease, timeout=timeout)
    #             await loop.run_in_executor(None, ff)
    #             self._lock_count = 1
    #         else:
    #             self._lock_count += 1
    #         try:
    #             yield
    #         finally:
    #             self._lock_count -= 1
    #             if self._lock_count <= 0:
    #                 await loop.run_in_executor(None, self._lease.release)
    #                 self._lease = None
    #                 self._lock_count = 0

    @ contextmanager
    def _provide_blob_client(self):
        if self._blob_client is None:
            bc = BlobClient(
                account_url=self._account_url,
                container_name=self._container_name,
                blob_name=self.blob_name,
                credential=self._sas_token or self._account_key,
            )
            self._blob_client = bc
            try:
                with bc:
                    yield
            finally:
                self._blob_client = None
        else:
            yield

    # TODO: how to optimize this part so that
    # new objects can reuse the ContainerClient?

    @ contextmanager
    def _provide_container_client(self):
        if self._container_client is None:
            cc = ContainerClient(
                account_url=self._account_url,
                container_name=self._container_name,
                credential=self._sas_token or self._account_key,
            )
            self._container_client = cc
            try:
                with cc:
                    yield
            finally:
                self._container_client = None
        else:
            yield

    @overrides
    def read_bytes(self) -> bytes:
        with self._provide_blob_client():
            try:
                return self._blob_client.download_blob().readall()
            except ResourceNotFoundError as e:
                raise FileNotFoundError(self) from e

    @overrides
    def remove_file(self):
        with self._provide_blob_client():
            try:

                self._blob_client.delete_blob(
                    delete_snapshots='include',
                    lease=self._lease)
            except ResourceNotFoundError:
                raise FileNotFoundError(self)

    @overrides
    def riterdir(self):
        with self._provide_container_client():
            prefix = self.blob_name + '/'
            k = len(prefix)
            for p in self._container_client.list_blobs(
                    name_starts_with=prefix):
                yield self / p.name[k:]

    @overrides
    def with_path(self, *paths):
        return self.__class__(
            *paths,
            account_name=self._account_name,
            account_key=self._account_key,
            sas_token=self._sas_token,
            container_name=self._container_name,
        )

    @overrides
    def write_bytes(self, data: bytes, *, overwrite=False) -> None:
        if self._path == '/':
            raise UnsupportedOperation(
                "can not write to root as a blob", self)

        with self._provide_blob_client():
            try:
                self._blob_client.upload_blob(
                    data,
                    overwrite=overwrite,
                    lease=self._lease)
            except ResourceExistsError as e:
                raise FileExistsError(self) from e



import concurrent.futures
import pathlib
import time
import pytest
from upathlib import Upath, LocalUpath, LockAcquisitionTimeoutError

# User runs these tests with a `Upath` object of their subclass,
# in a path that is safe for testing.


def test_basic(p: Upath):
    pp = p / '/abc/def/'
    assert pp.path == pathlib.PurePosixPath('/abc/def')
    print(repr(pp))

    pp = pp / 'x/y/z'
    assert pp.path == pathlib.PurePosixPath('/abc/def/x/y/z')
    print(repr(pp))

    pp /= 'xy/z'
    assert str(pp.path) == '/abc/def/x/y/z/xy/z'
    assert pp._path == str(pp.path)
    pp /= '..'
    assert pp._path == '/abc/def/x/y/z/xy'
    pp.joinpath('..')._path == '/abc/def/x/y/z'
    pp.joinpath('..', '..', '..', '..', '..')._path == '/'


def test_joinpath(path: Upath):
    try:
        pp = path.joinpath('/abc/def/', 'x/y') / 'ab.txt'
        assert str(pp.path) == '/abc/def/x/y/ab.txt'

        pp = pp.joinpath('../a/b.txt')
        assert pp == path / '/abc/def' / 'x/y/a/b.txt'
        assert pp.name == 'b.txt'
        assert pp.suffix == '.txt'

        p = pp

        pp = pp / '../../../../'
        assert str(pp.path) == '/abc/def'

        pp = p.joinpath('a', '.', 'b/c.data')
        assert str(pp.path) == '/abc/def/x/y/a/b.txt/a/b/c.data'
    except Exception:
        print('')
        print('repr:  ', repr(pp))
        print('str:   ', str(pp))
        print('path:  ', pp.path)
        print('_path: ', pp._path)
        raise


def test_compare(p: Upath):
    assert p.joinpath('abc/def') / 'x/y/z' == p / 'abc/def/x/y' / 'z'
    assert p / 'abc/def' < p.joinpath('abc/def', 'x')
    assert p.joinpath('abc/def/x', 'y/z') > p.joinpath('abc/def', 'x/y')


def test_read_write_rm_navigate(p: Upath):
    init_path = p._path

    p.rmrf()

    p1 = p / 'abc.txt'
    assert not p1.exists()
    p1.write_text('abc')

    assert p1.is_file()
    assert not p1.is_dir()

    assert p1.exists()
    assert p1.read_text() == 'abc'

    with pytest.raises(FileExistsError):
        p1.write_text('abcd')

    p1.write_json({'data': 'abcd'}, overwrite=True)  # type: ignore
    assert p1.read_json() == {'data': 'abcd'}  # type: ignore

    p /= 'a'
    assert p._path == f'{init_path}/a'
    assert not p.is_file()
    assert not p.is_dir()
    assert not p.exists()

    time.sleep(0.3)
    # wait to ensure the next file as diff timestamp
    # from the first file.

    p2 = p.joinpath('x.data')
    p2.write_bytes(b'x')
    p /= '..'
    assert p._path == init_path
    assert p2 == p.joinpath('a', 'x.data')
    assert p2.read_bytes() == b'x'

    p3 = p / 'a'

    assert p.ls() == [p3, p1]

    assert sorted(p.riterdir()) == [p2, p1]
    assert p3.file_info() is None
    fi1 = p1.file_info()
    fi2 = p2.file_info()
    print('')
    print('p1:', fi1)
    print('p2:', fi2)
    print('')
    assert fi1.mtime < fi2.mtime  # type: ignore
    print('file 1 size:', fi1.size)  # type: ignore
    print('file 2 size:', fi2.size)  # type: ignore
    assert fi1.size > fi2.size  # type: ignore

    assert p3.is_dir()
    assert p3.remove_dir() == 1
    assert not p3.exists()
    assert not p2.exists()

    assert p3.remove_dir() == 0
    assert p.ls() == [p1]
    assert p1.remove_dir() == 0
    p1.remove_file()
    with pytest.raises(FileNotFoundError):
        p1.remove_file()
    assert p.rmrf() == 0


def test_import_export(p: Upath):
    source = p
    source.rmrf()

    target = LocalUpath('/tmp/upath-test-target')
    target.rmrf()

    source_file = source / 'testfile'
    source_file.write_text('abc', overwrite=True)

    target.import_file(source_file)
    assert target.read_text() == 'abc'

    with pytest.raises(NotADirectoryError):
        # cant' write to `target/'samplefile'`
        # because `target` is a file.
        target.joinpath('samplefile').import_dir(source)

    target.rmrf()
    p2 = target.joinpath('samplefile')
    p2.import_dir(source)
    p3 = p2 / source_file.name
    assert target.ls() == [p2]
    assert p2.ls() == [p3]
    assert p3.read_text() == 'abc'

    p1 = source / 'a' / 'b' / 'c'
    assert p2.export_dir(p1) == 1
    p4 = p1 / source_file.name
    assert p4.read_text() == 'abc'

    assert p2.export_dir(source / 'a' / 'b') == 1
    assert (source / 'a' / 'b' / source_file.name).read_text() == 'abc'


def test_rename(p: Upath):
    p.rmrf()

    (p / 'a/a.txt').write_text('a')
    (p / 'b/b.txt').write_text('b')
    (p / 'c/d/e.txt').write_text('e')
    (p / 'c/d.txt').write_text('d')

    assert (p / 'a/a.txt').read_text() == 'a'

    p.joinpath('a/a.txt').rename_file('b/a.txt')
    assert not (p / 'a/a.txt').exists()
    assert (p / 'a/b/a.txt').read_text() == 'a'

    pp = (p / 'c').rename_dir('a/c')

    assert (pp / 'd/e.txt').read_text() == 'e'
    assert (pp / 'd.txt').read_text() == 'd'
    assert not (p / 'c').exists()


def _access_in_mp(root: Upath, path: str, timeout):
    p = root / path
    t0 = time.perf_counter()
    try:
        with p.lock(timeout=timeout):
            return time.perf_counter() - t0
    except LockAcquisitionTimeoutError:
        return t0 - time.perf_counter()


def test_lock(p: Upath, timeout=None, wait=3):
    p.rmrf()
    pp = p / 'testlock'
    with pp.lock(timeout=timeout):
        with concurrent.futures.ProcessPoolExecutor(1) as pool:
            t = pool.submit(_access_in_mp, p / '/', pp._path, wait)
            z = t.result()
            print('mp returned after', z, 'seconds')
            assert z <= -wait


def test_all(p: Upath):
    test_basic(p)
    test_joinpath(p)
    test_compare(p)

    test_read_write_rm_navigate(p)
    test_import_export(p)
    test_rename(p)



import abc
import gc
import json
import pickle
import zlib
from typing import TypeVar

import orjson
import zstandard


T = TypeVar('T')


MEGABYTE = 1048576  # 1024 * 1024
ZLIB_LEVEL = 3  # official default is 6
ZSTD_LEVEL = 3  # official default is 3


def _loads(func, data, **kwargs):
    if len(data) < MEGABYTE:
        return func(data, **kwargs)
    isgc = gc.isenabled()
    if isgc:
        gc.disable()
    try:
        return func(data, **kwargs)
    finally:
        if isgc:
            gc.enable()


class ByteSerializer(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def serialize(cls, x: T, **kwargs) -> bytes:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, y: bytes, **kwargs) -> T:
        raise NotImplementedError


class TextSerializer(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def serialize(cls, x: T, **kwargs) -> str:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, y: str, **kwargs) -> T:
        raise NotImplementedError


class PickleSerializer(ByteSerializer):
    @classmethod
    def serialize(cls, x, *, protocol=None):
        return pickle.dumps(x, protocol=protocol)

    @classmethod
    def deserialize(cls, y):
        return _loads(pickle.loads, y)


class ZPickleSerializer(PickleSerializer):
    @classmethod
    def serialize(cls, x, *, level=ZLIB_LEVEL, protocol=None):
        y = super().serialize(x, protocol=protocol)
        return zlib.compress(y, level=level)

    @classmethod
    def deserialize(cls, y):
        y = zlib.decompress(y)
        return super().deserialize(y)


class ZstdPickleSerializer(PickleSerializer):
    @classmethod
    def serialize(cls, x, *, level=ZSTD_LEVEL, protocol=None):
        y = super().serialize(x, protocol=protocol)
        return zstandard.compress(y, level=level)

    @classmethod
    def deserialize(cls, y):
        y = zstandard.decompress(y)
        return super().deserialize(y)


class JsonSerializer(TextSerializer):
    @classmethod
    def serialize(cls, x, **kwargs):
        return json.dumps(x, **kwargs)

    @classmethod
    def deserialize(cls, y, **kwargs):
        return _loads(json.loads, y, **kwargs)


class ZJsonSerializer(ByteSerializer):
    @classmethod
    def serialize(cls, x, *, level=ZLIB_LEVEL, **kwargs):
        y = json.dumps(x, **kwargs).encode()
        return zlib.compress(y, level=level)

    @classmethod
    def deserialize(cls, y, **kwargs):
        y = zlib.decompress(y).decode()
        return _loads(json.loads, y, **kwargs)


class ZstdJsonSerializer(ByteSerializer):
    @classmethod
    def serialize(cls, x, *, level=ZSTD_LEVEL, **kwargs):
        y = json.dumps(x, **kwargs).encode()
        return zstandard.compress(y, level=level)

    @classmethod
    def deserialize(cls, y, **kwargs):
        y = zstandard.decompress(y).decode()
        return _loads(json.loads, y, **kwargs)


class OrjsonSerializer(ByteSerializer):
    @classmethod
    def serialize(cls, x, **kwargs):
        return orjson.dumps(x, **kwargs)  # pylint: disable=no-member

    @classmethod
    def deserialize(cls, y):
        return _loads(orjson.loads, y)  # pylint: disable=no-member


class ZOrjsonSerializer(OrjsonSerializer):
    @classmethod
    def serialize(cls, x, *, level=ZLIB_LEVEL, **kwargs):
        y = super().serialize(x, **kwargs)
        return zlib.compress(y, level=level)

    @classmethod
    def deserialize(cls, y):
        y = zlib.decompress(y)
        return super().deserialize(y)


class ZstdOrjsonSerializer(OrjsonSerializer):
    @classmethod
    def serialize(cls, x, *, level=ZSTD_LEVEL, **kwargs):
        y = super().serialize(x, **kwargs)
        return zstandard.compress(y, level=level)

    @classmethod
    def deserialize(cls, y):
        y = zstandard.decompress(y)
        return super().deserialize(y)



import contextlib
import datetime
import logging
import os
import os.path
import pathlib
import shutil

import filelock
# `filelock` is also called `py-filelock`.
# Tried `fasteners` also. In one use case,
# `filelock` worked whereas `fasteners.InterprocessLock` failed.
#
# Other options to look into include
# `oslo.concurrency`, `pylocker`, `portalocker`.
from overrides import overrides

from ._upath import Upath, LockAcquisitionTimeoutError, FileInfo


logging.getLogger('filelock').setLevel(logging.WARNING)

logger = logging.getLogger(__name__)


class LocalUpath(Upath):
    def __init__(self, *pathsegments: str):
        assert os.name == 'posix'
        if pathsegments:
            parts = [str(pathlib.Path(*pathsegments).absolute())]
        else:
            parts = [str(pathlib.Path.cwd().absolute())]
        super().__init__(*parts)

    @overrides
    def _copy_file(self, target, *, overwrite=False):
        if not overwrite and target.is_file():
            raise FileExistsError(target)
        os.makedirs(target.localpath.parent, exist_ok=True)
        shutil.copyfile(self.localpath, target.localpath)
        # If target already exists, it will be overwritten.

    @overrides
    def export_dir(self, target: Upath, *, overwrite=False) -> int:
        if isinstance(target, LocalUpath):
            return super().export_dir(target, overwrite=overwrite)
        # `target` is a cloud store; it might have implemented
        # efficient 'download' functionality.
        return target.import_dir(self, overwrite=overwrite)

    @overrides
    def _export_file(self, target: Upath, *, overwrite=False):
        if isinstance(target, LocalUpath):
            return self._copy_file(target, overwrite=overwrite)
        # `target` is a cloud store; it might have implemented
        # efficient 'upload' functionality.
        target._import_file(self, overwrite=overwrite)

    @overrides
    def file_info(self):
        if not self.is_file():
            return None
        st = self.localpath.stat()
        return FileInfo(
            ctime=st.st_ctime,
            mtime=st.st_mtime,
            time_created=datetime.datetime.fromtimestamp(st.st_ctime),
            time_modified=datetime.datetime.fromtimestamp(st.st_mtime),
            size=st.st_size,
            details=st,
        )
        # If an existing file is written to again using `write_...`,
        # then its `ctime` and `mtime` are both updated.
        # My experiments showed that `ctime` and `mtime` are equal.

    @overrides
    def import_dir(self, source: Upath, *, overwrite=False) -> int:
        if isinstance(source, LocalUpath):
            return super().import_dir(source, overwrite=overwrite)
        return source.export_dir(self, overwrite=overwrite)

    @overrides
    def _import_file(self, source: Upath, *, overwrite=False):
        if isinstance(source, LocalUpath):
            return source._copy_file(self, overwrite=overwrite)
        # Call the other side in case it implements an efficient
        # file download.
        source._export_file(self, overwrite=overwrite)

    @overrides
    def is_dir(self) -> bool:
        return self.localpath.is_dir()

    @overrides
    def is_file(self) -> bool:
        return self.localpath.is_file()

    @overrides
    def iterdir(self):
        try:
            for p in self.localpath.iterdir():
                yield self / p.name
        except (NotADirectoryError, FileNotFoundError):
            pass

    @property
    def localpath(self) -> pathlib.Path:
        return pathlib.Path(self._path)

    @contextlib.contextmanager
    @overrides
    def lock(self, *, timeout=None):
        os.makedirs(self.localpath.parent, exist_ok=True)
        lock = filelock.FileLock(str(self.localpath))
        try:
            lock.acquire(timeout=timeout)
            yield
        except filelock.Timeout as e:
            raise LockAcquisitionTimeoutError(str(self)) from e
        finally:
            lock.release()

    @overrides
    def read_bytes(self) -> bytes:
        try:
            return self.localpath.read_bytes()
        except (IsADirectoryError, FileNotFoundError) as e:
            raise FileNotFoundError(self) from e

    @overrides
    def remove_dir(self) -> int:
        n = super().remove_dir()

        def _remove_empty_dir(path):
            for p in path.iterdir():
                assert p.is_dir()
                _remove_empty_dir(p)
            path.rmdir()

        if self.is_dir():
            _remove_empty_dir(self.localpath)
        return n

    @overrides
    def remove_file(self) -> None:
        self.localpath.unlink()

    @overrides
    def rename_dir(self, target, *, overwrite=False):
        target_ = super().rename_dir(target, overwrite=overwrite)

        def _remove_empty_dir(path):
            k = 0
            for p in path.iterdir():
                if p.is_dir():
                    k += _remove_empty_dir(p)
                else:
                    k += 1
            if k == 0:
                path.rmdir()
            return k

        _remove_empty_dir(self.localpath)

        return target_

    @overrides
    def _rename_file(self, target: str, *, overwrite=False):
        target = self.parent / target
        if not overwrite and target.is_file():
            raise FileExistsError(target)
        os.makedirs(target.localpath.parent, exist_ok=True)
        self.localpath.rename(target.localpath)

    @overrides
    def riterdir(self):
        for p in self.iterdir():
            if p.is_file():
                yield p
            elif p.is_dir():
                yield from p.riterdir()

    @overrides
    def write_bytes(self, data: bytes, *, overwrite=False):
        if self.is_file():
            if not overwrite:
                raise FileExistsError(self)
        elif self.is_dir():
            raise IsADirectoryError(self)
        else:
            self.parent.localpath.mkdir(exist_ok=True, parents=True)
        self.localpath.write_bytes(data)



__version__ = '0.6.4'

from ._upath import Upath, FileInfo, LockAcquisitionTimeoutError
from ._local import LocalUpath
from ._blob import BlobUpath


__all__ = [
    'Upath', 'LocalUpath', 'BlobUpath',
    'FileInfo',
    'LockAcquisitionTimeoutError',
]
